/*
 * Copyright (C) 2003-2004 Adam Geitgey <adam@rootnode.org>
 * Copyright (C) 2003 Hans Karlsson <karlsson.h@home.se>
 */


#include <Python.h>

#include <kdebug.h>
#include <qdir.h>
#include <qwidgetlist.h>
#include <krun.h>
#include <kparts/componentfactory.h>
#include <kparts/part.h>
#include <kwin.h>

#include "richtextlabel.h"

#include "xpm.h"
#include "karamba.h"
#include "themelistwindow.h"

#include "meter_python.h"
#include "bar_python.h"
#include "textlabel_python.h"
#include "richtextlabel_python.h"
#include "imagelabel_python.h"
#include "widget_python.h"
#include "menu_python.h"
#include "config_python.h"
#include "task_python.h"
#include "systray_python.h"
#include "misc_python.h"

// Menu IDs
#define EDITSCRIPT 1
#define THEMECONF  2

/*******************************************
 * Python methods are defined here.
 *   Each method accessible from python should have:
 *     - A wrapper function that returns a PyObject or appropriate python type
 *     - A C++ implementation of the python method, named the same as the python call
 *     - An entry in the python methods array so the call is accessible from python
 *
 *   Example:
 *     py_move_systay - wrapper function
 *     moveSystray - actual implementation of method
 *     {"moveSystray", py_move_systray, METH_VARARGS, "Move the Systray"} - array entry
 */

static PyMethodDef karamba_methods[] = {
    // Bar - bar_python.cpp
    {"createBar", py_createBar, METH_VARARGS, "Create new Bar."},
    {"deleteBar", py_deleteBar, METH_VARARGS, "Delete Bar."},
    {"getThemeBar", py_getThemeBar, METH_VARARGS, "Get Bar from .theme using it's name."},
    {"getBarSize", py_getBarSize, METH_VARARGS, "Get Bar size."},
    {"resizeBar", py_resizeBar, METH_VARARGS, "Resize Bar."},
    {"getBarPos", py_getBarPos, METH_VARARGS, "Get Bar position."},
    {"moveBar", py_moveBar, METH_VARARGS, "Move Bar."},
    {"hideBar", py_hideBar, METH_VARARGS, "Hide Bar."},
    {"showBar", py_showBar, METH_VARARGS, "Show Bar."},
    {"getBarSensor", py_getBarSensor, METH_VARARGS, "Get Bar sensor."},
    {"setBarSensor", py_setBarSensor, METH_VARARGS, "Set Bar sensor."},
    {"setBarImage", py_setBarImage, METH_VARARGS, "Set bar image"},
    {"getBarImage", py_getBarImage, METH_VARARGS, "Get bar image"},
    {"setBarVertical", py_setBarVertical, METH_VARARGS, "Set bar orientation"},
    {"getBarVertical", py_getBarVertical, METH_VARARGS, "Get bar orientation"},
    {"setBarValue", py_setBarValue, METH_VARARGS, "Set bar value"},
    {"getBarValue", py_getBarValue, METH_VARARGS, "Get bar value"},
    {"setBarMinMax", py_setBarMinMax, METH_VARARGS, "Set bar min & max"},
    {"getBarMinMax", py_getBarMinMax, METH_VARARGS, "Get bar min & max"},

    // TextLabel - textlabel_python.cpp
    {"createText", py_createText, METH_VARARGS, "Create new Text."},
    {"deleteText", py_deleteText, METH_VARARGS, "Delete Text."},
    {"getThemeText", py_getThemeText, METH_VARARGS, "Get Text from .theme using it's name."},
    {"getTextSize", py_getTextSize, METH_VARARGS, "Get Text size."},
    {"resizeText", py_resizeText, METH_VARARGS, "Resize Text."},
    {"getTextPos", py_getTextPos, METH_VARARGS, "Get Text position."},
    {"moveText", py_moveText, METH_VARARGS, "Move Text."},
    {"hideText", py_hideText, METH_VARARGS, "Hide Text."},
    {"showText", py_showText, METH_VARARGS, "Show Text."},
    {"getTextSensor", py_getTextSensor, METH_VARARGS, "Get Text sensor."},
    {"setTextSensor", py_setTextSensor, METH_VARARGS, "Set Text sensor."},
    {"changeText", py_setTextValue, METH_VARARGS, "Change a Text Sensor's Text"},
    {"getTextValue", py_getTextValue, METH_VARARGS, "Get Text value"},
    {"changeTextShadow", py_setTextShadow, METH_VARARGS, "Change a Text Shadow size"},
    {"getTextShadow", py_getTextShadow, METH_VARARGS, "Get a Text Shadow size"},
    {"changeTextFont", py_setTextFont, METH_VARARGS, "Change a Text Sensor's Font"},
    {"getTextFont", py_getTextFont, METH_VARARGS, "Get a Text Sensor's Font"},
    {"changeTextColor", py_setTextColor, METH_VARARGS, "Change a Text Sensor's Color"},
    {"getTextColor", py_getTextColor, METH_VARARGS, "Get a Text Sensor's Color"},
    {"changeTextSize", py_setTextFontSize, METH_VARARGS, "Change a Text Sensor's Font Size"},
    {"getTextFontSize", py_getTextFontSize, METH_VARARGS, "Get a Text Sensor's Font Size"},
    {"getTextAlign", py_getTextAlign, METH_VARARGS, "Get Text alignment."},
    {"setTextAlign", py_setTextAlign, METH_VARARGS, "Set Text alignment."},

    // RichTextLabel - richtextlabel_python.cpp
    {"createRichText", py_createRichText, METH_VARARGS, "Create a Rich Text Sensor"},
    {"deleteRichText", py_deleteRichText, METH_VARARGS, "Deletes a Rich Text Sensor"},
    {"getThemeRichText", py_getThemeRichText, METH_VARARGS, "Get Rich Text from .theme using it's name."},
    {"getRichTextSize", py_getRichTextSize, METH_VARARGS, "Get the (width, height) of a Rich Text Sensor"},
    {"resizeRichText", py_resizeRichText, METH_VARARGS, "Resize Rich Text."},
    {"setRichTextWidth", py_set_rich_text_width, METH_VARARGS, "Sets the width of a Rich Text Sensor"},
    {"getRichTextPos", py_getRichTextPos, METH_VARARGS, "Get Rich Text position."},
    {"moveRichText", py_moveRichText, METH_VARARGS, "Moves a Rich Text Sensor"},
    {"hideRichText", py_hideRichText, METH_VARARGS, "hides a Rich Text Sensor"},
    {"showRichText", py_showRichText, METH_VARARGS, "shows a Rich Text Sensor"},
    {"getRichTextSensor", py_getRichTextSensor, METH_VARARGS, "Get Rich Text sensor."},
    {"setRichTextSensor", py_setRichTextSensor, METH_VARARGS, "Set Rich Text sensor."},
    {"changeRichText", py_setRichTextValue, METH_VARARGS, "Change the content of a Rich Text Sensor"},
    {"getRichTextValue", py_getRichTextValue, METH_VARARGS, "Get Rich Text value"},
    {"changeRichTextFont", py_setRichTextFont, METH_VARARGS, "Change a Rich Text Sensor's Font"},
    {"getRichTextFont", py_getRichTextFont, METH_VARARGS, "Get a Rich Text Sensor's Font"},
    {"changeRichTextSize", py_setRichTextFontSize, METH_VARARGS, "Change a Rich Text Sensor's Font Size"},
    {"getRichTextFontSize", py_getRichTextFontSize, METH_VARARGS, "Get a Rich Text Sensor's Font Size"},
        
    // ImageLabel - imagelabel_python.cpp
    {"createImage", py_createImage, METH_VARARGS, "Create an Image"},
    {"createTaskIcon", py_createTaskIcon, METH_VARARGS, "Create an Image of the Icon for a Task"},
    {"createBackgroundImage", py_createBackgroundImage, METH_VARARGS, "Create an Image (only redraw it when background changes)"},
    {"deleteImage", py_deleteImage, METH_VARARGS, "Delete an Image"},
    {"getThemeImage", py_getThemeImage, METH_VARARGS, "Get image meter from .theme using it's name"},
    {"getImageSize", py_getImageSize, METH_VARARGS, "Get Image size."},
    {"getImageWidth", py_getImageWidth, METH_VARARGS, "Get the width of an Image"},
    {"getImageHeight", py_getImageHeight, METH_VARARGS, "Get the height of an Image"},
    {"getImagePos", py_getImagePos, METH_VARARGS, "Get Image position."},
    {"moveImage", py_moveImage, METH_VARARGS, "Move an Image"},
    {"hideImage", py_hideImage, METH_VARARGS, "Hide an Image"},
    {"showImage", py_showImage, METH_VARARGS, "Show an Image"},
    {"getImagePath", py_getImageValue, METH_VARARGS, "Get Image path."},
    {"setImagePath", py_setImageValue, METH_VARARGS, "Set Image path."},
    {"getImageSensor", py_getImageSensor, METH_VARARGS, "Get Image sensor."},
    {"setImageSensor", py_setImageSensor, METH_VARARGS, "Set Image sensor."},
    {"addImageTooltip", py_addImageTooltip, METH_VARARGS, "Create a Tooltip for an Image"},
    {"resizeImage", py_resizeImage, METH_VARARGS, "Scale an Image"},
    {"resizeImageSmooth", py_resizeImageSmooth, METH_VARARGS, "Scale an Image (slower, better looking)"},
    {"rotateImage", py_rotateImage, METH_VARARGS, "Rotate an Image"},
    {"removeImageTransformations", py_removeImageTransformations, METH_VARARGS, "Restore original size and orientation of an Image"},
    {"removeImageEffects", py_removeImageEffects, METH_VARARGS, "Remove Effects of an Image"},
    {"changeImageIntensity", py_changeImageIntensity, METH_VARARGS, "Change Intensity of an Image"},
    {"changeImageChannelIntensity", py_changeImageChannelIntensity, METH_VARARGS, "Change Intensity of an Image Channel"},
    {"changeImageToGray", py_changeImageToGray, METH_VARARGS, "Converts an Image to Grayscale"},

    // Menu - menu_python.cpp
    {"createMenu", py_create_menu, METH_VARARGS, "Create a popup menu"},
    {"deleteMenu", py_delete_menu, METH_VARARGS, "Delete a popup menu"},
    {"addMenuItem", py_add_menu_item, METH_VARARGS, "Add a popup menu entry"},
    {"addMenuSeparator", py_add_menu_separator, METH_VARARGS, "Add a popup menu seperator item"},
    {"removeMenuItem", py_remove_menu_item, METH_VARARGS, "Remove a popup menu entry"},
    {"popupMenu", py_popup_menu, METH_VARARGS, "Popup a menu at a specified location"},
    
    // Config - config_python.cpp
    {"addMenuConfigOption", py_add_menu_config_option, METH_VARARGS, "Add a configuration entry to the menu"},
    {"setMenuConfigOption", py_set_menu_config_option, METH_VARARGS, "Set a configuration entry in the menu"},
    {"readMenuConfigOption", py_read_menu_config_option, METH_VARARGS, "Read a configuration entry in the menu"},
    {"readConfigEntry", py_read_config_entry, METH_VARARGS, "Read a configuration entry"},
    {"writeConfigEntry", py_write_config_entry, METH_VARARGS, "Writes a configuration entry"},
    
    // Widget - widget_python.cpp
    {"moveWidget", py_move_widget, METH_VARARGS, "Move Widget to x,y"},
    {"resizeWidget", py_resize_widget, METH_VARARGS, "Resize Widget to width,height"},
    {"createWidgetMask", py_create_widget_mask, METH_VARARGS, "Create a clipping mask for this widget"},
    {"redrawWidget", py_redraw_widget, METH_VARARGS, "Update Widget to reflect your changes"},
    {"redrawWidgetBackground", py_redraw_widget_background, METH_VARARGS, "Update Widget to reflect background image changes"},
    {"getWidgetPosition", py_get_widget_position, METH_VARARGS, "Get Widget Position"},

    // Task - task_python.cpp
    {"getStartupList", py_get_startup_list, METH_VARARGS, "Get the system startup list"},
    {"getStartupInfo", py_get_startup_info, METH_VARARGS, "Get all the info for a startup"},
    {"getTaskList", py_get_task_list, METH_VARARGS, "Get the system task list"},
    {"getTaskNames", py_get_task_names, METH_VARARGS, "Get the system task list in name form"},
    {"getTaskInfo", py_get_task_info, METH_VARARGS, "Get all the info for a task"},
    {"performTaskAction", py_perform_task_action, METH_VARARGS, "Do something with a task, such as minimize it"},
    
    // System Tray - systray_python.cpp
    {"createSystray", py_create_systray, METH_VARARGS, "Create a Systray"},
    {"hideSystray", py_hide_systray, METH_VARARGS, "Hide the Systray"},
    {"showSystray", py_show_systray, METH_VARARGS, "Show the Systray"},
    {"moveSystray", py_move_systray, METH_VARARGS, "Move the Systray"},
    
    // Misc - misc_python.cpp
    {"getThemePath", py_get_theme_path, METH_VARARGS,  "Get the file path of the theme"},
    {"openTheme", py_open_theme, METH_VARARGS, "Open a new theme"},
    {"acceptDrops", py_accept_drops, METH_VARARGS, "Allows widget to receive Drop (I.E. Drag and Drop) events"},
    {"toggleShowDesktop", py_toggle_show_desktop, METH_VARARGS, "Show/Hide the desktop"},
    {"execute", py_execute_command, METH_VARARGS, "Execute a command"},
    {"executeInteractive", py_execute_command_interactive, METH_VARARGS, "Execute a command and get it's output (stdout)"},
    {"attachClickArea", (PyCFunction)py_attach_clickArea, METH_VARARGS|METH_KEYWORDS, "Add a clickArea to the given text or image"},
    {"createClickArea", py_create_click_area, METH_VARARGS, "Create a Click Area Sensor"},
    {NULL, NULL, 0 ,NULL}
};



karamba::karamba(QString fn, bool reloading)
        :  QWidget( 0,"karamba", Qt::WGroupLeader | WStyle_Customize |  WRepaintNoErase| WStyle_NoBorder | WDestructiveClose  )
{

    char pypath[1024];
    themePath = QFileInfo( fn ).dirPath(true) + "/";
    themeName = QFileInfo( fn ).baseName();
    themeFile = fn;
   
    QFile themeConfigFile(QDir::home().absPath() + "/.superkaramba/" + themeName + ".rc");
    // Tests if config file Exists
    if (!QFileInfo(themeConfigFile).exists())
    {
       // Create config file
       themeConfigFile.open(IO_ReadWrite);
       themeConfigFile.close();
    }

    // Creates KConfig Object
    config = new KConfig(QFileInfo(themeConfigFile).filePath(), false, false);
    config -> sync();
    config -> setGroup("internal");

    connect(qApp,SIGNAL(lastWindowClosed()),qApp,SLOT(quit()));

    pythonThemeExtensionLoaded = false;

    //Add theme path to python path so that we can find the python file
    snprintf(pypath, 1023, "sys.path.insert(0, '%s')", themePath.ascii());

    // get the global python thread lock
    PyEval_AcquireLock();

    // get a reference to the PyInterpreterState
    extern PyThreadState * mainThreadState;
    PyInterpreterState * mainInterpreterState = mainThreadState->interp;

    // create a thread state object for this thread
    PyThreadState * myThreadState = PyThreadState_New(mainInterpreterState);
    PyThreadState_Swap(myThreadState);

    // load the .py file for this .theme
    PyRun_SimpleString("import sys");
    PyRun_SimpleString(pypath);
    PyRun_SimpleString("sys.path.insert(0, '')");


    PyImport_AddModule("karamba");
    Py_InitModule("karamba", karamba_methods);


    pName = PyString_FromString(themeName.ascii());

    pModule = PyImport_Import(pName);

    //Make sure the module is up to date.
    if (reloading)
      PyImport_ReloadModule(pModule);

    if (pModule != NULL) {
	pDict = PyModule_GetDict(pModule);
	if (pDict != NULL) {
	  pythonThemeExtensionLoaded = true;
	}	
	//HERE	
    }
     else {
       PyErr_Print();
       fprintf(stderr, "------------------------------------------------------\n");       
       fprintf(stderr, "What does ImportError mean?\n");       
       fprintf(stderr, "\n");       
       fprintf(stderr, "It means that I couldn't load a python add-on for %s.theme\n", themeName.ascii());       
       fprintf(stderr, "If this is a regular theme and doesn't use python\n");       
       fprintf(stderr, "extensions, then nothing is wrong.\n");       
       fprintf(stderr, "------------------------------------------------------\n");       
       
     }

    //Clean up this thread's python interpreter reference
    PyThreadState_Swap(NULL);
    PyThreadState_Clear(myThreadState);
    PyThreadState_Delete(myThreadState);
    PyEval_ReleaseLock();
    //This thread's python refence is now gone.  A new reference must be created for each thread.


    //Matthew Kay: set window type to "dock" (plays better with taskbar themes this way)
    KWin::setType(winId(), NET::Dock);
#ifdef HAVE_KDE_32
    //KDE 3.2 addition for the always on top issues
    KWin::setState(winId(), NET::KeepBelow);
#endif

    widgetMask = 0;
    info = new NETWinInfo( qt_xdisplay(), winId(), qt_xrootwin(), NET::WMState );

    // could be replaced with TaskManager
    kWinModule = new KWinModule();
    desktop = 0;

        connect( kWinModule,SIGNAL(currentDesktopChanged(int)), this,SLOT(currentDesktopChanged(int)) );

    // Setup of the Task Manager Callbacks
    connect(&taskManager, SIGNAL(activeTaskChanged(Task*)), this, SLOT(activeTaskChanged(Task*)) );
    connect(&taskManager, SIGNAL(taskAdded(Task*)), this, SLOT(taskAdded(Task*)) );
    connect(&taskManager, SIGNAL(taskRemoved(Task*)), this, SLOT(taskRemoved(Task*)) );
    connect(&taskManager, SIGNAL(startupAdded(Startup*)), this, SLOT(startupAdded(Startup*)) );
    connect(&taskManager, SIGNAL(startupRemoved(Startup*)), this, SLOT(startupRemoved(Startup*)) );

    themeConfMenu = new KPopupMenu( this);
    themeConfMenu -> setCheckable(true);

    /* XXX - need to be able to delete all these DesktopChangeSlot objects */
    DesktopChangeSlot *dslot;
    
    int mid;

    toDesktopMenu = new KPopupMenu (this);
    toDesktopMenu -> setCheckable(true);
    mid = toDesktopMenu -> insertItem ("&All Desktops", 
                                 dslot = new DesktopChangeSlot(this,0), 
                                 SLOT(receive()));
    dslot->setMenuId(mid);

    toDesktopMenu -> insertSeparator();
    for (int ndesktop=1; ndesktop <= kWinModule->numberOfDesktops(); ndesktop++)
    {
      QString name = "Desktop &";
      name += ('0' + ndesktop);
      
      mid = toDesktopMenu -> insertItem (name, dslot = new DesktopChangeSlot(this, ndesktop), SLOT(receive()));
      dslot->setMenuId(mid);
    }
    

    kpop = new KPopupMenu( this );
    kpop -> setCheckable(true);

    keditpop = new KPopupMenu ( this );

    keditpop -> insertItem(SmallIconSet("edit"),tr("Edit &theme"),this,SLOT(editConfig()), CTRL+Key_E );
    keditpop -> insertItem(SmallIconSet("edit"),tr("Edit &script"),this,SLOT(editScript()), ALT+Key_E, EDITSCRIPT);

    // Test if Theme Script exists
    QFileInfo scriptFile(themePath + "/" + themeName + ".py");
    keditpop -> setItemEnabled(EDITSCRIPT, scriptFile.exists());

    accColl = new KActionCollection( this );
    menuAccColl = new KActionCollection( this );
    
    kpop->insertItem( SmallIconSet("reload"),tr("Update"), this, SLOT(updateSensors()), Key_F5 );
    toggleLocked = new KToggleAction (  tr("Toggle &locked position"), SmallIconSet("locked"),
                   CTRL+Key_L, this, SLOT( slotToggleLocked() ), accColl  , "Locked position" );
    accColl->insert(toggleLocked);
    toggleLocked -> setChecked(true);

    toggleLocked->plug(kpop);

    toggleFastTransforms = new KToggleAction(tr("Use &fast image scaling"),
       CTRL+Key_F, this, SLOT( slotToggleFastTransforms() ), accColl, "Fast transformations");

    accColl->insert(toggleFastTransforms);
    toggleFastTransforms -> setChecked(true);

    toggleFastTransforms -> plug(kpop);

    kpop->insertSeparator();

    kpop->insertItem( SmallIconSet("fileopen"),tr("&Open new theme"), this, SLOT(startNewKaramba()), CTRL+Key_O );
    kpop->insertItem(tr("&Edit"), keditpop);
    kpop->insertItem(tr("Configure &theme"), themeConfMenu, THEMECONF);
    kpop->setItemEnabled(THEMECONF, false);
    kpop->insertItem(tr("To Des&ktop"), toDesktopMenu);

    kpop->insertItem( SmallIconSet("reload3"),tr("&Reload theme"),this,SLOT(reloadConfig()), CTRL+Key_R );
    kpop->insertItem( SmallIconSet("fileclose"),tr("&Close this theme"), this, SLOT(killWidget()), CTRL+Key_C );

    kpop->insertSeparator();
    kpop->insertItem( SmallIconSet("find"),tr("&Download more themes"), this, SLOT(downloadThemes()), CTRL+Key_D );

    kpop->insertSeparator();

    kpop->insertItem( SmallIconSet("exit"),tr("&Quit"), this, SLOT(quitKaramba()), CTRL+Key_Q );

    kpop->polish();

    numberOfConfMenuItems = 0;

    systray = 0;
    foundKaramba = false;
    onTop = false;
    managed = false;
    fixedPosition = false;
    defaultTextField = new TextField();

    meterList = new QObjectList();
    meterList->setAutoDelete( true );
    sensorList = new QObjectList();
    sensorList->setAutoDelete( true );
    clickList = new QObjectList();
    timeList = new QObjectList();
    imageList = new QObjectList();
    menuList = new QObjectList();
    menuList->setAutoDelete( true );

    client = kapp->dcopClient();
    if (!client->isAttached())
        client->attach();
    appId = client->registerAs(qApp->name());


    this->setBackgroundMode( NoBackground);
    if( !(onTop || managed))
        this->lower();


    if( !parseConfig() )
    {
        setFixedSize(0,0);
        QTimer::singleShot( 100, this, SLOT(killWidget()) );

        qDebug("Could not read config file.");
    }
    else
    {
        kroot = new KarambaRootPixmap((QWidget*)this);
        kroot->start();


    }
    
    // Karamba specific Config Entries
    if (!config -> readBoolEntry("lockedPosition", true))
    {
        toggleLocked -> setChecked(false);
        slotToggleLocked();
    }

    if (!config -> readBoolEntry("fastTransforms", true))
    {
        toggleFastTransforms -> setChecked(false);
        slotToggleFastTransforms();
    }

    desktop = config -> readNumEntry("desktop", desktop);
    if (desktop > kWinModule->numberOfDesktops())
    {
        desktop = 0;
    }

    if (desktop)
    {
        info->setDesktop( desktop );
    }
    else
        info->setDesktop( NETWinInfo::OnAllDesktops);

    // Read Themespecific Config Entries
    config -> setGroup("theme");
    if (config -> hasKey("widgetPosX") && config -> hasKey("widgetPosY"))
    {
      int xpos = config -> readNumEntry("widgetPosX");
      int ypos = config -> readNumEntry("widgetPosY");
      
      if (xpos < 0)
        xpos = 0;
      if (ypos < 0)
        ypos = 0;

       move(xpos, ypos);
    }
    
    haveUpdated = 0;
    this->setMouseTracking(true);


    setFocusPolicy(QWidget::StrongFocus);

    //Add self to list of open themes
    QWidgetList  *list = QApplication::allWidgets();
    QWidgetListIt it( *list );         // iterate over the widgets
    QWidget * w;
    while ( (w=it.current()) != 0 ) {  // for each widget...
      ++it;
      if (QString(w->name()) == "ThemeListWindow") {
	ThemeListWindow* k = (ThemeListWindow*) w;
	k->addTheme(this);
      } else {
      }
    }
    delete list;                      // delete the list, not the widgets
}

karamba::~karamba()
{
   //Remove self from list of open themes
    QWidgetList  *list = QApplication::allWidgets();
    QWidgetListIt it( *list );         // iterate over the widgets
    QWidget * w;
    while ( (w=it.current()) != 0 ) {  // for each widget...
      ++it;
      if (QString(w->name()) == "ThemeListWindow") {
	ThemeListWindow* k = (ThemeListWindow*) w;
	k->removeTheme(this);
      } else {
      }
    }
    delete list;                      // delete the list, not the widgets


  widgetClosed();
    writeConfigData();
    delete config;

    if(meterList != 0)
    {
        meterList->clear();
        delete meterList;
    }

    if( sensorList != 0 )
    {
        sensorList->clear();
        delete sensorList;
    }

    if( imageList != 0 )
    {
        imageList->clear();
        delete imageList;
    }
	
	if( clickList != 0 )
    {
        clickList->clear();
        delete clickList;
    }
	
	if( timeList != 0 )
    {
        timeList->clear();
        delete timeList;
    }

    if( toggleLocked != 0)
        delete toggleLocked;
    if( accColl != 0)
        delete accColl;
    if ( menuAccColl != 0)
	delete menuAccColl;
    
    if( keditpop != 0)
      delete keditpop;

    if( themeConfMenu != 0)
      delete themeConfMenu;

    if( kpop != 0)
        delete kpop;

    if( widgetMask != 0)
      delete widgetMask;

    delete kWinModule;
    delete defaultTextField;

    //Clean up Python references
    if (pythonThemeExtensionLoaded) {
      // get the global python thread lock
      PyEval_AcquireLock();
      
      // get a reference to the PyInterpreterState
      extern PyThreadState * mainThreadState;
      PyInterpreterState * mainInterpreterState = mainThreadState->interp;
      
      // create a thread state object for this thread
      PyThreadState * myThreadState = PyThreadState_New(mainInterpreterState);
      PyThreadState_Swap(myThreadState);
      
      //Displose of current python module so we can reload in constructor.
      Py_DECREF(pModule);
      Py_DECREF(pName);
      
      //Clean up this thread's python interpreter reference
      PyThreadState_Swap(NULL);
      PyThreadState_Clear(myThreadState);
      PyThreadState_Delete(myThreadState);
      PyEval_ReleaseLock();
      //This thread's python refence is now gone.  A new reference must be created for each thread.
    }
}

bool karamba::parseConfig()
{
    QTimer *m_sysTimer = new QTimer(this);

    QFile file(themeFile);
    QString line;
    QString meter;
    int interval = 0;

    if ( file.open(IO_ReadOnly | IO_Translate) )
    {
        QTextStream t( &file );        // use a text stream
        QValueStack<QPoint> offsetStack;
        offsetStack.push(QPoint(0,0));

        int x=0;
        int y=0;
        int w=0;
        int h=0;
        while( ( line = t.readLine() ) !=0 )
        {
            QRegExp rx("^\\s*(\\S+)");
            rx.search(line);
            meter = rx.cap(1).upper();

            x = getInt("X",line)+offsetStack.top().x();
            y = getInt("Y",line)+offsetStack.top().y();
            w = getInt("W",line);
            h = getInt("H",line);


            if( meter == "KARAMBA" && !foundKaramba )
            {
	      //                qDebug("karamba");
                toggleLocked->setChecked( getBoolean( "LOCKED", line ) );
                slotToggleLocked();


                getInt( "X",line );
                getInt( "Y",line );

                x = ( x < 0 ) ? 0:x;
                y = ( y < 0 ) ? 0:y;
                move(x,y);


                if( w == 0 ||  h == 0)
                {
                    w = 300;
                    h = 300;
                }
                setFixedSize(w,h);

                if( getBoolean( "RIGHT", line ) )
                {
                    QDesktopWidget *d = QApplication::desktop();
                    x = d->width() - w;
                }
                else
                    if( getBoolean( "LEFT", line ) )
                    {
                        x = 0;
                    }

                if( getBoolean( "BOTTOM", line ) )
                {
                    QDesktopWidget *d = QApplication::desktop();
                    y = d->height() - h;
                }
                else
                    if( getBoolean( "TOP", line ) )
                    {
                        y = 0;
                    }

                pm = QPixmap(size());

                if( getBoolean("ONTOP", line ) )
                {
                    onTop = true;
                    KWin::setState( winId(), NET::StaysOnTop );
                }

                if( getBoolean("MANAGED", line ) ) {
		  managed = true;
		  reparent(0, Qt::WType_Dialog | WStyle_Customize | WStyle_NormalBorder|  WRepaintNoErase | WDestructiveClose, pos());
		    
		} else {
		  info->setState( NETWinInfo::SkipTaskbar | NETWinInfo::SkipPager,NETWinInfo::SkipTaskbar | NETWinInfo::SkipPager );
		  if (onTop) {
                    KWin::setState( winId(), NET::StaysOnTop );

		  }
		}

                if (getBoolean("ONALLDESKTOPS", line))
                {
                    desktop = 200; // ugly
                }


                bool dfound=false;
                int desktop = getInt("DESKTOP", line, dfound);
                if (dfound)
                {
                    info->setDesktop( dfound );
                }
                if( getBoolean("TOPBAR", line ) )
                {
                    move(x,0);
                    KWin::setStrut( winId(), 0, 0, h, 0 );
                    toggleLocked->setChecked( true );
                    slotToggleLocked();
                    toggleLocked->setEnabled(false);

                }

                if( getBoolean("BOTTOMBAR", line ) )
                {
                    int dh = QApplication::desktop()->height();
                    move( x, dh - h );
                    KWin::setStrut( winId(), 0, 0, 0, h );
                    toggleLocked->setChecked( true );
                    slotToggleLocked();
                    toggleLocked->setEnabled(false);


                }

                if( getBoolean("RIGHTBAR", line ) )
                {
                   int dw = QApplication::desktop()->width();
                   move( dw - w, y );
                   KWin::setStrut( winId(), 0, w, 0, 0 );
                   toggleLocked->setChecked( true );
                   slotToggleLocked();
                   toggleLocked->setEnabled(false);

                }
		if( getBoolean("LEFTBAR", line) )
		{
		   int dw = QApplication::desktop()->width();
		   move( 0, y );
		   KWin::setStrut( winId(), w, 0, 0, 0 );
		   toggleLocked->setChecked( true );
		   slotToggleLocked();
		   toggleLocked->setEnabled(false);

		}

		bool found = false;
                QString path = getString( "MASK", line, found );

                if( 1 )
                {		  

		  QFileInfo info(path);
		  if( info.isRelative())
                    path = themePath +"/" + path;

		  widgetMask = new QBitmap(path);
  
		  setMask(*widgetMask);
		  
                }

                interval = getInt("INTERVAL",line);
                interval = ( interval == 0 ) ? 5000 : interval;
                foundKaramba = true;
            }

            if( meter == "THEME" )
            {
                QString path = getString( "PATH", line );
                QFileInfo info(path);
                if( info.isRelative())
                    path = themePath +"/" + path;
		//                qDebug("path: "+path);
                (new karamba( path, false ))->show();
            }

            if( meter == "<GROUP>" )
            {
                int offsetX = offsetStack.top().x();
                int offsetY = offsetStack.top().y();
                offsetStack.push( QPoint( offsetX + getInt("X",line),
                                          offsetY + getInt("Y",line) ) );
            }

            if( meter == "</GROUP>" )
            {
                offsetStack.pop();
            }

            if( meter == "CLICKAREA" )
            {
                if( !hasMouseTracking() )
                    setMouseTracking(true);
                ClickArea *tmp = new ClickArea( x, y, w, h );
                tmp->setOnClick( getString( "ONCLICK", line ) );

                setSensor( line, (Meter*) tmp );
                clickList->append( tmp );
                if( getBoolean( "PREVIEW", line ) )
                    meterList->append( tmp );
            }

            // program sensor without a meter
            if(meter == "SENSOR=PROGRAM")
            {
                setSensor(line, 0 );
            }

            if(meter == "IMAGE")
            {
                QString file = getString("PATH",line);
                QString file_roll = getString("PATHROLL",line);
                int xon = getInt("XROLL",line);
                int yon = getInt("YROLL",line);
                QString tiptext = getString("TOOLTIP",line);
                QString name = getString("NAME",line);
                bool bg = getBoolean("BACKGROUND", line);
                xon = ( xon <= 0 ) ? x:xon;
                yon = ( yon <= 0 ) ? y:yon;

                ImageLabel *tmp = new ImageLabel(x,y,0,0);
                tmp->setThemePath( themePath );
                tmp->setValue( file );
                tmp->parseImages(file, file_roll, x,y, xon, yon);
                tmp->setKaramba(this);
                tmp->setBackground(bg);
                if (name != "")
                  tmp->setName(name.ascii());
                if (!tiptext.isEmpty())
                  tmp->setTooltip(tiptext);

                setSensor(line, (Meter*) tmp );
                meterList->append (tmp );
                imageList->append (tmp );
            }

            if( meter == "DEFAULTFONT" )
            {
                int r,g,b;

                delete defaultTextField;
                defaultTextField = new TextField( );

                get3Int( "COLOR", line, r, g, b );
                defaultTextField->setColor(r, g, b);
                get3Int( "BGCOLOR", line, r, g, b );
                defaultTextField->setBGColor( r, g, b );

                defaultTextField->setFont( getString( "FONT", line ) );
                defaultTextField->setFontSize( getInt( "FONTSIZE", line ) );
                defaultTextField->setAlignment( getString( "ALIGN", line ) );
                defaultTextField->setFixedPitch( getBoolean( "FIXEDPITCH", line ) );
                defaultTextField->setShadow( getInt( "SHADOW", line ) );

            }


            if(meter == "TEXT" || meter == "CLICKMAP" || meter == "RICHTEXT")
            {
                int r,g,b;

                TextField* tmpText = new TextField();
                bool paramFound = false;

                get3Int( "COLOR", line, r, g, b, paramFound );
                if ( paramFound || !defaultTextField ) {
                    tmpText->setColor ( r, g, b );
		}
                else
                    if ( defaultTextField )
                    {
                        defaultTextField->getColor().rgb ( &r, &g, &b );
                        tmpText->setColor( r, g, b );
                    }

                get3Int( "BGCOLOR", line, r, g, b, paramFound );
                if ( paramFound  || !defaultTextField )
                    tmpText->setBGColor( r, g, b );
                else
                    if ( defaultTextField )
                    {
                        defaultTextField->getBGColor().rgb ( &r, &g, &b );
                        tmpText->setBGColor( r, g, b );
                    }

                QString dFont = getString("FONT", line, paramFound);
                if( paramFound || !defaultTextField )
                    tmpText->setFont( dFont );
                else
                    if( defaultTextField )
                        tmpText->setFont( defaultTextField->getFont() );


                int dFontSize = getInt( "FONTSIZE", line, paramFound );
                if( paramFound || !defaultTextField )
                    tmpText->setFontSize( dFontSize );
                else
                    if( defaultTextField )
                        tmpText->setFontSize( defaultTextField->getFontSize() );

                QString dAlign = getString( "ALIGN", line, paramFound );
                if( paramFound || !defaultTextField )
                    tmpText->setAlignment( dAlign );
                else
                    if( defaultTextField )
                        tmpText->setAlignment( defaultTextField->getAlignment() );

                bool dFp = getBoolean( "FIXEDPITCH", line, paramFound );
                if( paramFound || !defaultTextField )
                    tmpText->setFixedPitch( dFp );
                else
                    if( defaultTextField )
                        tmpText->setFixedPitch( defaultTextField->getFixedPitch() );

                int dShadow = getInt( "SHADOW", line, paramFound );
                if( paramFound || !defaultTextField )
                    tmpText->setShadow( dShadow );
                else
                    if( defaultTextField )
                        tmpText->setShadow( defaultTextField->getShadow() );

                // ////////////////////////////////////////////////////
                // Now handle the specifics
                if( meter == "TEXT" )
                {
                    
                    TextLabel *tmp = new TextLabel( x, y, w, h );
                    tmp->setTextProps( tmpText );
                    tmp->setValue( getString( "VALUE", line ) );

                    QString name = getString( "NAME", line);
                    if (name != "")
                        tmp->setName(name.ascii());

                    setSensor( line, (Meter*)tmp );
                    meterList->append ( tmp );
                }

                if(meter == "CLICKMAP")
                {
                    if( !hasMouseTracking() )
                        setMouseTracking(true);
                    ClickMap *tmp = new ClickMap( x, y, w, h );
                    tmp->setTextProps( tmpText );

                    setSensor( line, (Meter*) tmp );
                    // set all params
                    clickList -> append(tmp);
                    meterList->append( tmp );

                }
                
                if( meter == "RICHTEXT" )
                {
                    RichTextLabel *tmp = new RichTextLabel( this, x, y, w, h );

                    bool dUl = getBoolean( "UNDERLINE", line, paramFound );
                    if( !paramFound )
                        dUl = false;

                    tmp->setText( getString( "VALUE", line ), dUl );
                    tmp->setTextProps( tmpText );
                    tmp->setThemePath( themePath );
                    tmp->setWidth(w);
                    tmp->setHeight(h);

                    QString name = getString( "NAME", line);
                    if (name != "")
                        tmp->setName(name.ascii());

                    setSensor( line, (Meter*)tmp );
                    clickList -> append(tmp);
                    meterList->append ( tmp );
                }
            }

            if(meter == "BAR")
            {
                Bar *tmp =
                    new Bar( x, y, w, h );
                tmp->setThemePath( themePath );
                tmp->setImage( getString( "PATH", line ) );
                tmp->setVertical( getBoolean( "VERTICAL", line ) );
                bool maxFound = false;
                int max = getInt( "MAX", line, maxFound );
                if( maxFound )
                    tmp->setMax( max );
                bool minFound = false;
                int min = getInt( "MIN", line, minFound );
                if( minFound )
                    tmp->setMin( min );
                tmp->setValue( getInt( "VALUE", line ) );
                QString name = getString("NAME",line);
                if (name != "")
                    tmp->setName(name.ascii());
                setSensor( line, (Meter*)tmp );
                meterList->append ( tmp );
            }

            if(meter == "GRAPH")
            {
                int r, g, b;
                int points = getInt("POINTS",line);
                get3Int("COLOR", line, r, g, b);

                Graph *tmp = new Graph( x, y, w, h, points );
                bool maxFound = false;
                int max = getInt( "MAX", line, maxFound );
                if( maxFound )
                    tmp->setMax( max );
                bool minFound = false;
                int min = getInt( "MIN", line, minFound );
                if( minFound )
                    tmp->setMin( min );

                tmp->setColor( r , g, b );

                setSensor( line, (Graph*)tmp );
                meterList->append ( tmp );
            }
        }
    }
    if( !foundKaramba )
    {
        //  interval = initKaramba( "", 0, 0, 0, 0 );
        //   this->close(true);
        //delete this;
        return false;
    }
    else
    {
        connect(m_sysTimer, SIGNAL(timeout()), SLOT(step()));

        interval = interval == 0 ? 1000 : interval;
        m_sysTimer->start(interval);

	//Start the widget running
        QTimer::singleShot( 0, this, SLOT(step()) );

        if( !(onTop || managed) )
            lowerTimer.start();

        return true;
    }
}

void karamba::startNewKaramba()
{
    QStringList fileNames;
    fileNames = KFileDialog::getOpenFileNames(QString::null, "*.theme *.ctheme|Themes", 0, "Open Themes");
    for ( QStringList::Iterator it = fileNames.begin(); it != fileNames.end(); ++it )
    {
        QFileInfo file( *it );
        if( file.exists() )
        {
            (new karamba( *it, false ))->show();
        }
    }
}

void karamba::popupNotify(int item)
{
  
}

void karamba::reloadConfig()
{
    writeConfigData();
    QFileInfo file( themeFile );
    //qDebug(themeFile);
    if( file.exists() )
    {
        (new karamba( themeFile, true ))->show();
    }
    killWidget();
}

void karamba::killWidget() {


  close();
}

void karamba::quitKaramba()
{
	//qWarning("Quit Karamba.");
	qApp -> closeAllWindows();

	qApp -> quit();
}


void karamba::editConfig()
{
    QFileInfo fileInfo( themeFile );
    QString path;

    if( fileInfo.isRelative() )
    {
        path = themePath + "/" + themeFile;
    }
    else
    {
        path = themeFile;
    }
    KRun::runURL( KURL( path ), "text/plain" );

}

void karamba::editScript()
{
    QFileInfo fileInfo( themeFile );
    QString path;

    if( fileInfo.isRelative() )
    {
        path = themePath + "/" + themeName + ".py";
    }
    else
    {
        path = QFileInfo(themeFile).dirPath() + "/" + themeName + ".py";
    }
    KRun::runURL( KURL( path ), "text/plain" );

}


int karamba::getInt(QString w, QString &line )
{
    bool b;
    return getInt( w, line, b );
}

int karamba::getInt(QString w, QString &line, bool &found )
{
    QRegExp rx( "\\W+" + w +"=([-]?\\d+)", false );
    found = (rx.search(line)==-1)?false:true;
    return rx.cap(1).toInt();
}

void karamba::get3Int(QString w, QString &line, int &a, int &b, int &c )
{
    bool bo;
    get3Int ( w, line, a, b, c, bo );
}

void karamba::get3Int(QString w, QString &line, int &a, int &b, int &c, bool &found )
{
    QRegExp rx( "\\W+" + w + "=([-]?\\d+),([-]?\\d+),([-]?\\d+)", false );
    found = (rx.search(line)==-1)?false:true;
    a = rx.cap(1).toInt();
    b = rx.cap(2).toInt();
    c = rx.cap(3).toInt();
}

QString karamba::getString(QString w, QString &line)
{
    bool b;
    return getString ( w, line, b );
}

QString karamba::getString(QString w, QString &line, bool &found)
{
    QRegExp rx( "\\W+" + w + "=\"([^\"]*)\"", false );
    found = (rx.search(line)==-1)?false:true;
    if (rx.cap(1).isEmpty())
    {
        rx = QRegExp(w + "=(\\S+)", false);
        rx.search(line);
        return rx.cap(1);
    }
    else
    {
        return rx.cap(1);
    }

}

bool karamba::getBoolean( QString w, QString &line )
{
    bool b;
    return getBoolean ( w, line, b );
}

bool karamba::getBoolean( QString w, QString &line, bool &found )
{
    return  ( getString( w, line, found ).lower() == "true")?true:false;
}

QString karamba::findSensorFromMap(Sensor* sensor)
{
  QMap<QString,Sensor*>::Iterator it;

  for ( it = sensorMap.begin(); it != sensorMap.end(); ++it )
    {
        if (it.data() == sensor)
          return it.key();
    }
  return "";
}

Sensor* karamba::findSensorFromList(Meter* meter)
{
  QObjectListIt it( *sensorList ); // iterate over meters

  while ( it != 0 )
    {
        if (((Sensor*) *it)->hasMeter(meter))
            return ((Sensor*)*it);
        ++it;
    }
  return NULL;
}

QString karamba::getSensor(Meter* meter)
{
  QString s;
  Sensor* sensor = findSensorFromList(meter);
  if (sensor)
    s = findSensorFromMap(sensor);
  return s;
}

void karamba::deleteMeterFromSensors(Meter* meter)
{
  Sensor* sensor = findSensorFromList(meter);

  if (sensor)
    {
        sensor->deleteMeter(meter);
        if (sensor->isEmpty())
          {
            QString s = findSensorFromMap(sensor);
            sensorMap.erase(s);
            sensorList->removeRef(sensor);
          }
    }
}

void karamba::setSensor(QString &line, Meter* meter)
{
    Sensor* sensor = 0;
    
    deleteMeterFromSensors(meter);

    QString sens = getString("SENSOR",line).upper();

    if( sens == "CPU" )
    {
        QString cpuNbr = getString( "CPU", line );
        sensor = sensorMap["CPU"+cpuNbr];
        if (sensor == 0)
        {
            int interval = getInt("INTERVAL",line);
            interval = (interval == 0)?1000:interval;
            sensor = ( sensorMap["CPU"+cpuNbr] = new CPUSensor( cpuNbr, interval ) );
            sensorList->append( sensor );
        }
        SensorParams *sp = new SensorParams(meter);
        sp->addParam("FORMAT",getString("FORMAT",line));
        sp->addParam("DECIMALS",getString("DECIMALS",line));

        sensor->addMeter(sp);
        sensor->setMaxValue(sp);

    }

    if( sens == "MEMORY" )
    {
        sensor = sensorMap["MEMORY"];
        if (sensor == 0)
        {
            int interval = getInt("INTERVAL",line);
            interval = (interval == 0)?3000:interval;
            sensor = ( sensorMap["MEMORY"] = new MemSensor( interval ) );
            sensorList->append( sensor );
        }
        SensorParams *sp = new SensorParams(meter);
        sp->addParam("FORMAT",getString("FORMAT",line));

        sensor->addMeter(sp);
        sensor->setMaxValue(sp);
    }


    if( sens == "DISK" )
    {
        sensor = sensorMap["DISK"];
        if (sensor == 0)
        {
            int interval = getInt("INTERVAL",line);
            interval = (interval == 0)?5000:interval;
            sensor = ( sensorMap["DISK"] = new DiskSensor( interval ) );
            sensorList->append( sensor );
        }
        // meter->setMax( ((DiskSensor*)sensor)->getTotalSpace(mntPt)/1024 );
        SensorParams *sp = new SensorParams(meter);
        QString mntPt = getString("MOUNTPOINT",line);
        if( mntPt == ""  )
            mntPt = "/";

        sp->addParam("MOUNTPOINT",mntPt);
        sp->addParam("FORMAT",getString("FORMAT",line));
        sensor->addMeter(sp);
        sensor->setMaxValue(sp);
    }

    if( sens == "NETWORK")
    {
        int interval = getInt("INTERVAL",line );
        interval = (interval == 0)?2000:interval;
        QString device = getString("DEVICE", line );
        sensor = sensorMap["NETWORK"+device];
        if (sensor == 0)
        {
            sensor = ( sensorMap["NETWORK"+device] = new NetworkSensor( device, interval ) );
            sensorList->append( sensor );
        }
        SensorParams *sp = new SensorParams(meter);
        sp->addParam( "FORMAT", getString( "FORMAT", line ) );
        sp->addParam( "DECIMALS", getString( "DECIMALS", line ) );
        sensor->addMeter(sp);
    }

    if( sens == "UPTIME" )
    {
        sensor = sensorMap["UPTIME"];
        if (sensor == 0)
        {
            int interval = getInt("INTERVAL",line);
            interval = (interval == 0)?60000:interval;
            sensor = ( sensorMap["UPTIME"] = new UptimeSensor( interval ));
            sensorList->append( sensor );

        }
        SensorParams *sp = new SensorParams(meter);
        sp->addParam( "FORMAT", getString( "FORMAT", line ) );
        sensor->addMeter(sp);
    }

    if( sens == "SENSOR" )
    {
        sensor = sensorMap["SENSOR"];
        if (sensor == 0)
        {
            int interval = getInt("INTERVAL",line);
            interval = (interval == 0)?30000:interval;
            sensor = ( sensorMap["SENSOR"] = new SensorSensor( interval ) );
            sensorList->append( sensor );
        }
        SensorParams *sp = new SensorParams(meter);
        sp->addParam("FORMAT",getString("FORMAT",line));
        sp->addParam("TYPE",getString("TYPE",line));
        sensor->addMeter(sp);
    }


    if( sens == "TEXTFILE" )
    {
        QString path = getString( "PATH", line );
        bool rdf = getBoolean( "RDF", line );
        sensor = sensorMap["TEXTFILE"+path];
        if (sensor == 0)
        {
            int interval = getInt( "INTERVAL", line );
            interval = ( interval == 0 )?60000:interval;
            QString encoding = getString( "ENCODING", line );

            sensor = ( sensorMap["TEXTFILE"+path] =
                           new TextFileSensor( path, rdf, interval, encoding ) );
            sensorList->append( sensor );
        }
        SensorParams *sp = new SensorParams(meter);
        sp->addParam("LINE",QString::number(getInt("LINE",line)));
        sensor->addMeter(sp);
    }


    if( sens == "TIME")
    {
        sensor = sensorMap["DATE"];
        if (sensor == 0)
        {
            int interval = getInt("INTERVAL",line);
            interval = (interval == 0)?60000:interval;
            sensor = ( sensorMap["DATE"] = new DateSensor( interval ) );
            sensorList->append( sensor );
			timeList->append( sensor );
        }
        SensorParams *sp = new SensorParams(meter);
        sp->addParam("FORMAT",getString("FORMAT",line));
        sp->addParam("CALWIDTH",getString("CALWIDTH",line));
        sp->addParam("CALHEIGHT",getString("CALHEIGHT",line));
        sensor->addMeter(sp);
    }

#ifdef HAVE_XMMS

    if( sens == "XMMS" )
    {
        sensor = sensorMap["XMMS"];
        if (sensor == 0)
        {
            int interval = getInt("INTERVAL",line);
            interval = (interval == 0)?1000:interval;
            QString encoding = getString( "ENCODING", line );

            sensor = ( sensorMap["XMMS"] = new XMMSSensor( interval, encoding ) );
            sensorList->append( sensor );
        }
        SensorParams *sp = new SensorParams(meter);
        sp->addParam("FORMAT",getString("FORMAT",line));
        sensor->addMeter(sp);
        sensor->setMaxValue(sp);
    }
#endif // HAVE_XMMS


    if( sens == "NOATUN" )
    {
        sensor = sensorMap["NOATUN"];
        if (sensor == 0)
        {
            int interval = getInt("INTERVAL",line);
            interval = (interval == 0)?1000:interval;
            sensor = ( sensorMap["NOATUN"] = new NoatunSensor( interval, client ) );
            sensorList->append( sensor );
        }
        SensorParams *sp = new SensorParams(meter);
        sp->addParam("FORMAT",getString("FORMAT",line));
        sensor->addMeter(sp);
        sensor->setMaxValue(sp);
    }

    if( sens == "PROGRAM")
    {
        QString progName = getString("PROGRAM",line);
        sensor = sensorMap["PROGRAM"+progName];
        if (sensor == 0)
        {
            int interval = getInt("INTERVAL",line);
            interval = (interval == 0)?3600000:interval;
            QString encoding = getString( "ENCODING", line );

            sensor = (sensorMap["PROGRAM"+progName] =
                          new ProgramSensor( progName, interval, encoding ) );
            sensorList->append( sensor );
        }
        SensorParams *sp = new SensorParams(meter);
        sp->addParam( "LINE", QString::number( getInt( "LINE", line ) ) );
        sp->addParam( "THEMAPATH", themePath );
        sensor->addMeter(sp);
    }

    if( sens == "RSS" )
    {
        QString source = getString( "SOURCE", line );
        QString format = getString( "FORMAT", line );

        sensor = sensorMap["RSS"+source];
        if (sensor == 0)
        {
            int interval = getInt( "INTERVAL", line );
            interval = ( interval == 0 )?60000:interval;
            QString encoding = getString( "ENCODING", line );

            sensor = ( sensorMap["RSS"+source] =
                           new RssSensor( source, interval, format, encoding ) );
            sensorList->append( sensor );
        }
        SensorParams *sp = new SensorParams(meter);
        sp->addParam("SOURCE",getString("SOURCE",line));
        sensor->addMeter(sp);
    }

    if (sensor != 0)
    {
        QTimer::singleShot( 0, sensor, SLOT(update()) );
        sensor->start();
    }
}

void karamba::passMenuOptionChanged(QString key, bool value)
{
//Everything below is to call the python callback function

    if (pythonThemeExtensionLoaded) {

      // get the global lock
      PyEval_AcquireLock();
      // get a reference to the PyInterpreterState

      //Python thread references
      extern PyThreadState * mainThreadState;

      PyInterpreterState * mainInterpreterState = mainThreadState->interp;
      // create a thread state object for this thread
      PyThreadState * myThreadState = PyThreadState_New(mainInterpreterState);
      PyThreadState_Swap(myThreadState);

      pFunc = PyDict_GetItemString(pDict, "menuOptionChanged");

      if (pFunc && PyCallable_Check(pFunc)) {
	        pArgs = PyTuple_New(3);

          pValue = PyInt_FromLong((int)this);
          PyTuple_SetItem(pArgs, 0, pValue);

          pValue = PyString_FromString(key.ascii());
          PyTuple_SetItem(pArgs, 1, pValue);

          pValue = PyInt_FromLong((int)value);
          PyTuple_SetItem(pArgs, 2, pValue);

          pValue = PyObject_CallObject(pFunc, pArgs);

          if (pValue != NULL) {
            Py_DECREF(pValue);
          }
          else {
            qWarning("Call to menuOptionChanged failed");
	    PyErr_Print();
          }
          Py_DECREF(pArgs);
      }
      else {
	//callback wasn't found
      }


      // swap my thread state out of the interpreter
      PyThreadState_Swap(NULL);
      // clear out any cruft from thread state object
      PyThreadState_Clear(myThreadState);
      // delete my thread state object
      PyThreadState_Delete(myThreadState);
      // release the lock
      PyEval_ReleaseLock();
    }
}

void karamba::meterClicked(QMouseEvent* e, Meter* meter)
{
    //qWarning("karamba::meterClicked");
    if (pythonThemeExtensionLoaded && haveUpdated)
    {
        // get the global lock
        PyEval_AcquireLock();
        // get a reference to the PyInterpreterState

        //Python thread references
        extern PyThreadState * mainThreadState;

        PyInterpreterState * mainInterpreterState = mainThreadState -> interp;
        // create a thread state object for this thread
        PyThreadState * myThreadState = PyThreadState_New(mainInterpreterState);
        PyThreadState_Swap(myThreadState);

        pFunc = PyDict_GetItemString(pDict, "meterClicked");

        if (pFunc && PyCallable_Check(pFunc))
        {
	  //qWarning ("meterClicked is going to be called");

            pArgs = PyTuple_New(3);

            pValue = PyInt_FromLong((int)this);
            PyTuple_SetItem(pArgs, 0, pValue);

            // if meter is of type RichTextLabel
            if (RichTextLabel* richText = dynamic_cast<RichTextLabel*>(meter))
            {
                pValue = PyString_FromString(richText -> anchorAt(e -> x(), e -> y()).ascii());
            }
            else
            {
                pValue = PyInt_FromLong((int)meter);
            }
            PyTuple_SetItem(pArgs, 1, pValue);

            int button = 0;

            if( e->button() == Qt::LeftButton )
                button = 1;
            else if( e->button() == Qt::MidButton )
                button = 2;
            else if( e->button() == Qt::RightButton )
                button = 3;
            pValue = PyInt_FromLong(button);
            PyTuple_SetItem(pArgs, 2, pValue);


            pValue = PyObject_CallObject(pFunc, pArgs);
            if (pValue != NULL)
            {
                Py_DECREF(pValue);
            }
            else
            {
                qWarning("Call to meterClicked failed");
		PyErr_Print();
            }
            Py_DECREF(pArgs);
        }
        else
        {
	  //callback not found
        }

        // swap my thread state out of the interpreter
        PyThreadState_Swap(NULL);
        // clear out any cruft from thread state object
        PyThreadState_Clear(myThreadState);
        // delete my thread state object
        PyThreadState_Delete(myThreadState);
        // release the lock
        PyEval_ReleaseLock();
    }
}

void karamba::passClick( QMouseEvent *e )
{
	QObjectListIt it2( *timeList ); // iterate over meters
    while ( it2 != 0 )
    {
        (( DateSensor* ) *it2)->toggleCalendar( e );
        ++it2;
    }

    QObjectListIt it(*clickList);
    while (it != 0)
    {
        if (((Meter*)(*it)) -> click(e))
        {
            // callback
            meterClicked(e, (Meter*)(*it));
        }
        ++it;
    }

    //Everything below is to call the python callback function

    if (pythonThemeExtensionLoaded && haveUpdated) {

      // get the global lock
      PyEval_AcquireLock();
      // get a reference to the PyInterpreterState

      //Python thread references
      extern PyThreadState * mainThreadState;

      PyInterpreterState * mainInterpreterState = mainThreadState->interp;
      // create a thread state object for this thread
      PyThreadState * myThreadState = PyThreadState_New(mainInterpreterState);
      PyThreadState_Swap(myThreadState);

      pFunc = PyDict_GetItemString(pDict, "widgetClicked");

      if (pFunc && PyCallable_Check(pFunc)) {
	pArgs = PyTuple_New(4);

	pValue = PyInt_FromLong((int)this);
	PyTuple_SetItem(pArgs, 0, pValue);

	pValue = PyInt_FromLong(e->x());
	PyTuple_SetItem(pArgs, 1, pValue);

	pValue = PyInt_FromLong(e->y());
	PyTuple_SetItem(pArgs, 2, pValue);

	int button = 0;

	if( e->button() == Qt::LeftButton )
	  button = 1;
	else if( e->button() == Qt::MidButton )
	  button = 2;
	else if( e->button() == Qt::RightButton )
	  button = 3;

	pValue = PyInt_FromLong(button);
	PyTuple_SetItem(pArgs, 3, pValue);


	pValue = PyObject_CallObject(pFunc, pArgs);
	if (pValue != NULL) {
	  Py_DECREF(pValue);
	}
	else {
	  qWarning("Call to widgetClicked failed");
	  PyErr_Print();
	}
	Py_DECREF(pArgs);
      }
      else {
	//callback not found
      }


      // swap my thread state out of the interpreter
      PyThreadState_Swap(NULL);
      // clear out any cruft from thread state object
      PyThreadState_Clear(myThreadState);
      // delete my thread state object
      PyThreadState_Delete(myThreadState);
      // release the lock
      PyEval_ReleaseLock();
    }
}


void karamba::passWheelClick( QWheelEvent *e )
{
    //Everything below is to call the python callback function

    if (pythonThemeExtensionLoaded && haveUpdated) {

      // get the global lock
      PyEval_AcquireLock();
      // get a reference to the PyInterpreterState

      //Python thread references
      extern PyThreadState * mainThreadState;

      PyInterpreterState * mainInterpreterState = mainThreadState->interp;
      // create a thread state object for this thread
      PyThreadState * myThreadState = PyThreadState_New(mainInterpreterState);
      PyThreadState_Swap(myThreadState);


      //    PyRun_SimpleString("iconbar.yo()");

      pFunc = PyDict_GetItemString(pDict, "widgetClicked");

      if (pFunc && PyCallable_Check(pFunc)) {
	pArgs = PyTuple_New(4);

	pValue = PyInt_FromLong((int)this);
	PyTuple_SetItem(pArgs, 0, pValue);

	pValue = PyInt_FromLong(e->x());
	PyTuple_SetItem(pArgs, 1, pValue);

	pValue = PyInt_FromLong(e->y());
	PyTuple_SetItem(pArgs, 2, pValue);

	int button = 0;

	if( e->delta() > 0 )

	  button = 4;
	else
	  button = 5;



	pValue = PyInt_FromLong(button);
	PyTuple_SetItem(pArgs, 3, pValue);


	pValue = PyObject_CallObject(pFunc, pArgs);
	if (pValue != NULL) {
	  Py_DECREF(pValue);
	}
	else {
	  qWarning("Call to widgetClicked failed");
	  PyErr_Print();
	}
	Py_DECREF(pArgs);
      }
      else {
	//callback not found
      }


      // swap my thread state out of the interpreter
      PyThreadState_Swap(NULL);
      // clear out any cruft from thread state object
      PyThreadState_Clear(myThreadState);
      // delete my thread state object
      PyThreadState_Delete(myThreadState);
      // release the lock
      PyEval_ReleaseLock();
    }
}

void karamba::mousePressEvent( QMouseEvent *e )
{

    if( e->button() == RightButton )
    {

        kpop->exec(QCursor::pos());
    }
    else
    {
        clickPos = e->pos();
        if( toggleLocked -> isChecked() )
            passClick( e );
        if( !(onTop || managed))
            lower();
    }
}

void karamba::wheelEvent( QWheelEvent *e )
{
  passWheelClick( e );
}

void karamba::mouseReleaseEvent( QMouseEvent *e )
{
    clickPos = e->pos();
    step();
}

void karamba::mouseDoubleClickEvent( QMouseEvent *e )
{
    if( !toggleLocked -> isChecked() )
    {
        passClick( e );
    }
}


void karamba::keyPressEvent(QKeyEvent *e) 
{
  //Everything below is to call the python callback function

  if (pythonThemeExtensionLoaded) {

    // get the global lock
    PyEval_AcquireLock();
    // get a reference to the PyInterpreterState

    //Python thread references
    extern PyThreadState * mainThreadState;

    PyInterpreterState * mainInterpreterState = mainThreadState->interp;
    // create a thread state object for this thread
    PyThreadState * myThreadState = PyThreadState_New(mainInterpreterState);
    PyThreadState_Swap(myThreadState);

    pFunc = PyDict_GetItemString(pDict, "keyPressed");

    if (pFunc && PyCallable_Check(pFunc)) {
      pArgs = PyTuple_New(2);

      pValue = PyInt_FromLong((int)this);
      PyTuple_SetItem(pArgs, 0, pValue);

      pValue = PyString_FromString(e->text().ascii());
      PyTuple_SetItem(pArgs, 1, pValue);

      pValue = PyObject_CallObject(pFunc, pArgs);
      if (pValue != NULL) {
	Py_DECREF(pValue);
      }
      else {
	qWarning("Call to keyPressed failed");
	PyErr_Print();
      }
      Py_DECREF(pArgs);
      }
    else {
      //callback not found
    }


    // swap my thread state out of the interpreter
    PyThreadState_Swap(NULL);
    // clear out any cruft from thread state object
    PyThreadState_Clear(myThreadState);
    // delete my thread state object
    PyThreadState_Delete(myThreadState);
    // release the lock
    PyEval_ReleaseLock();
  }

  

}


void karamba::mouseMoveEvent( QMouseEvent *e )
{
    if( e->state() !=  0 && e->state() < 16 && !toggleLocked -> isChecked() )
    {
        move( e->globalPos() - clickPos );
    }
    else
    {
        // Change cursor over ClickArea
        QObjectListIt it(*clickList);
        bool insideArea = false;

        while (it != 0)
        {
            insideArea = ((Meter*)(*it)) -> insideActiveArea(e -> x(), e -> y());
            if (insideArea)
            {
                break;
            }
            ++it;
        }

        if(insideArea)
        {
            if( cursor().shape() != PointingHandCursor )
                setCursor( PointingHandCursor );
        }
        else
        {
            if( cursor().shape() != ArrowCursor )
                setCursor( ArrowCursor );
        }

        QObjectListIt image_it( *imageList);	// iterate over image sensors
        while ( image_it != 0 )
        {
            ((ImageLabel*) *image_it)->rolloverImage(e);
            ++image_it;
        }
    }


    if (pythonThemeExtensionLoaded)
    {
        // get the global lock
        PyEval_AcquireLock();
        // get a reference to the PyInterpreterState

        //Python thread references
        extern PyThreadState * mainThreadState;

        PyInterpreterState * mainInterpreterState = mainThreadState->interp;
        // create a thread state object for this thread
        PyThreadState * myThreadState = PyThreadState_New(mainInterpreterState);
        PyThreadState_Swap(myThreadState);

        pFunc = PyDict_GetItemString(pDict, "widgetMouseMoved");

        if (pFunc && PyCallable_Check(pFunc))
        {
            pArgs = PyTuple_New(4);

            pValue = PyInt_FromLong((int)this);
            PyTuple_SetItem(pArgs, 0, pValue);

            pValue = PyInt_FromLong(e->x());
            PyTuple_SetItem(pArgs, 1, pValue);

            pValue = PyInt_FromLong(e->y());
            PyTuple_SetItem(pArgs, 2, pValue);

            int button = 0;

	    //Modified by Ryan Nickell (p0z3r@mail.com) 03/16/2004
	    // This will work now, but only when you move at least
	    // one pixel in any direction with your mouse.
            //if( e->button() == Qt::LeftButton )
	    if( e->state() == LeftButton)
                button = 1;
            //else if( e->button() == Qt::MidButton )
	    else if( e->state() == MidButton )
                button = 2;
            //else if( e->button() == Qt::RightButton )
	    else if( e->state() == RightButton )
                button = 3;

            pValue = PyInt_FromLong(button);
            PyTuple_SetItem(pArgs, 3, pValue);

            pValue = PyObject_CallObject(pFunc, pArgs);
            if (pValue != NULL)
            {
                Py_DECREF(pValue);
            }
            else
            {
                //callback failed
            }
            Py_DECREF(pArgs);
        }
        else
        {
            //callback not found
        }

        // swap my thread state out of the interpreter
        PyThreadState_Swap(NULL);
        // clear out any cruft from thread state object
        PyThreadState_Clear(myThreadState);
        // delete my thread state object
        PyThreadState_Delete(myThreadState);
        // release the lock
        PyEval_ReleaseLock();
    }
}

void karamba::closeEvent ( QCloseEvent *  qc)
{
  qc->accept();

//  this->close(true);
//  delete this;
}

void karamba::paintEvent ( QPaintEvent *e)
{
    if( !(onTop || managed))
        if( lowerTimer.elapsed() > 100 )
        {
            this->lower();
            lowerTimer.restart();
        }
    QRect rect = e->rect();
    bitBlt(this,rect.topLeft(),&pm,rect,Qt::CopyROP);
}

void karamba::updateSensors()
{
    QObjectListIt it( *sensorList ); // iterate over meters
    while ( it != 0 )
    {
        ((Sensor*) *it)->update();
        ++it;
    }
    QTimer::singleShot( 500, this, SLOT(step()) );
}

void karamba::updateBackground(KSharedPixmap* kpm)
{
  background = QPixmap(*kpm);

  QPixmap buffer = QPixmap(size());

  pm = QPixmap(size());
  buffer.fill(Qt::black);

    QObjectListIt it( *imageList ); // iterate over meters
    p.begin(&buffer);
    bitBlt(&buffer,0,0,&background,0,Qt::CopyROP);

    while ( it != 0 )
    {
      if (((ImageLabel*) *it)->background == 1)
        ((ImageLabel*) *it)->mUpdate(&p, 1);
      ++it;
    }
    p.end();


    bitBlt(&pm,0,0,&buffer,0,Qt::CopyROP);
    background = pm;

    QPixmap buffer2 = QPixmap(size());

    pm = QPixmap(size());
    buffer2.fill(Qt::black);

    QObjectListIt it2( *meterList ); // iterate over meters
    p.begin(&buffer2);
    bitBlt(&buffer2,0,0,&background,0,Qt::CopyROP);

    while ( it2 != 0 )
    {
        ((Meter*) *it2)->mUpdate(&p);
        ++it2;
    }
    p.end();

    bitBlt(&pm,0,0,&buffer2,0,Qt::CopyROP);
    if (systray != 0){
      systray->updateBackgroundPixmap(buffer2);
    }
    repaint();

}

void karamba::currentDesktopChanged( int )
{
    kroot->repaint( true );
}


void karamba::externalStep()
{
  QPixmap buffer = QPixmap(size());

  pm = QPixmap(size());
  buffer.fill(Qt::black);

    QObjectListIt it( *meterList ); // iterate over meters
    p.begin(&buffer);
    bitBlt(&buffer,0,0,&background,0,Qt::CopyROP);

    while ( it != 0 )
    {
        ((Meter*) *it)->mUpdate(&p);
        ++it;
    }
    p.end();

    bitBlt(&pm,0,0,&buffer,0,Qt::CopyROP);
    repaint();
}


void karamba::step()
{
    pm = QPixmap(size());
    QPixmap buffer = QPixmap(size());
    buffer.fill(Qt::black);

    QObjectListIt it( *meterList ); // iterate over meters
    p.begin(&buffer);

    bitBlt(&buffer,0,0,&background,0,Qt::CopyROP);


    while ( it != 0 )
    {
        ((Meter*) *it)->mUpdate(&p);
        ++it;
    }
    p.end();

    bitBlt(&pm,0,0,&buffer,0,Qt::CopyROP);
    update();


    if (pythonThemeExtensionLoaded) {

      // get the global lock
      PyEval_AcquireLock();
      // get a reference to the PyInterpreterState

      //Python thread references
      extern PyThreadState * mainThreadState;

      PyInterpreterState * mainInterpreterState = mainThreadState->interp;
      // create a thread state object for this thread
      PyThreadState * myThreadState = PyThreadState_New(mainInterpreterState);
      PyThreadState_Swap(myThreadState);

      if (haveUpdated == 0) {
	pFunc = PyDict_GetItemString(pDict, "initWidget");
	haveUpdated = 1;
      }
      else
	pFunc = PyDict_GetItemString(pDict, "widgetUpdated");


      if (pFunc && PyCallable_Check(pFunc)) {
	pArgs = PyTuple_New(1);

	  pValue = PyInt_FromLong((int)this);
	  PyTuple_SetItem(pArgs, 0, pValue);
	
	pValue = PyObject_CallObject(pFunc, pArgs);
	if (pValue != NULL) {
	  Py_DECREF(pValue);
	}
	else {
	  qWarning("Call to widgetUpdated failed");
	  PyErr_Print();
	}
	Py_DECREF(pArgs);

      }
	else {
	  //callback not found
	}

      // swap my thread state out of the interpreter
      PyThreadState_Swap(NULL);
      // clear out any cruft from thread state object
      PyThreadState_Clear(myThreadState);
      // delete my thread state object
      PyThreadState_Delete(myThreadState);
      // release the lock
      PyEval_ReleaseLock();

    
    }
}

void karamba::widgetClosed() {
    if (pythonThemeExtensionLoaded) {

      // get the global lock
      PyEval_AcquireLock();
      // get a reference to the PyInterpreterState

      //Python thread references
      extern PyThreadState * mainThreadState;

      PyInterpreterState * mainInterpreterState = mainThreadState->interp;
      // create a thread state object for this thread
      PyThreadState * myThreadState = PyThreadState_New(mainInterpreterState);
      PyThreadState_Swap(myThreadState);

      pFunc = PyDict_GetItemString(pDict, "widgetClosed");


      if (pFunc && PyCallable_Check(pFunc)) {
	pArgs = PyTuple_New(1);

	  pValue = PyInt_FromLong((int)this);
	  PyTuple_SetItem(pArgs, 0, pValue);
	
	pValue = PyObject_CallObject(pFunc, pArgs);
	if (pValue != NULL) {
	  Py_DECREF(pValue);
	}
	else {
	  qWarning("Call to widgetClosed failed");
	  //	  PyErr_Print();
	}
	Py_DECREF(pArgs);

      }
	else {
	  //callback not found
	}

      // swap my thread state out of the interpreter
      PyThreadState_Swap(NULL);
      // clear out any cruft from thread state object
      PyThreadState_Clear(myThreadState);
      // delete my thread state object
      PyThreadState_Delete(myThreadState);
      // release the lock
      PyEval_ReleaseLock();

    
    }
}
void karamba::slotToggleLocked()
{
  //    qDebug("toggleLocked");
    bool locked = toggleLocked->isChecked();
    if( toggleLocked -> isChecked() )
    {
        toggleLocked->setIconSet(SmallIconSet("lock"));
    }
    else
    {
        toggleLocked->setIconSet(SmallIconSet("move"));
    }

}

void karamba::slotToggleFastTransforms()
{
//	bool fastTransforms = toggleFastTransforms -> isChecked();
// 	if (toggleFastTransforms -> isChecked())
// 	{
//     toggleFastTransforms -> setIconSet(SmallIconSet("ok"));
// 	}
// 	else
// 	{
//     QPixmap ok_disabled;
// 		toggleFastTransforms -> setIconSet(ok_disabled);
// 	}
  //config.setGroup("internal");
  //config.writeEntry("fastTransforms", toggleFastTransforms -> isChecked());
}


bool karamba::useSmoothTransforms()
{
  return !toggleFastTransforms -> isChecked();
}

void karamba::writeConfigData()
{
    config -> setGroup("internal");
    config -> writeEntry("lockedPosition", toggleLocked -> isChecked() );
    config -> writeEntry("fastTransforms", toggleFastTransforms -> isChecked() );
    config -> writeEntry("desktop", desktop );
    config -> setGroup("theme");
    // Widget Position
    config -> writeEntry("widgetPosX", x());
    config -> writeEntry("widgetPosY", y());

    // Widget Size
    config -> writeEntry("widgetWidth", width());
    config -> writeEntry("widgetHeight", height());

    // write changes to DiskSensor
    config -> sync();
    //qWarning("Config File ~/.superkaramba/%s.rc written.", themeName.ascii());
}

void karamba::slotToggleConfigOption(QString key, bool value)
{
    config -> setGroup("config menu");
    config -> writeEntry(key, value);
    passMenuOptionChanged(key, value);
}

void karamba::addMenuConfigOption(QString key, QString name)
{
    kpop -> setItemEnabled(THEMECONF, true);

    SignalBridge* action = new SignalBridge(this, key, menuAccColl);
    KToggleAction* confItem = new KToggleAction (name, KShortcut::null(), action, SLOT(receive()),
                                   menuAccColl, key);
    confItem -> setName(key);

    menuAccColl -> insert(confItem);

    connect(action, SIGNAL( enabled(QString, bool) ), this, SLOT( slotToggleConfigOption(QString, bool) ));

    config -> setGroup("config menu");
    confItem -> setChecked(config -> readBoolEntry(key));

    confItem -> plug(themeConfMenu);

    numberOfConfMenuItems++;
}

bool karamba::setMenuConfigOption(QString key, bool value)
{
    KToggleAction* menuAction = ((KToggleAction*)menuAccColl -> action(key));
    if (menuAction == NULL)
    {
	    qWarning("Menu action %s not found.", key.ascii());
	    return false;
    }
    else
    {
	    menuAction -> setChecked(value);
	    return true;
    }
}

bool karamba::readMenuConfigOption(QString key)
{
	KToggleAction* menuAction = ((KToggleAction*)menuAccColl -> action(key));
	if (menuAction == NULL)
	{
		qWarning("Menu action %s not found.", key.ascii());
		return false;
	}
	else
	{
		return menuAction -> isChecked();
	}
}

void karamba::passMenuItemClicked(int id)
{
    //Everything below is to call the python callback function

    if (pythonThemeExtensionLoaded)
    {
        // get the global lock
        PyEval_AcquireLock();
        // get a reference to the PyInterpreterState

        //Python thread references
        extern PyThreadState * mainThreadState;

        PyInterpreterState * mainInterpreterState = mainThreadState->interp;
        // create a thread state object for this thread
        PyThreadState * myThreadState = PyThreadState_New(mainInterpreterState);
        PyThreadState_Swap(myThreadState);

        pFunc = PyDict_GetItemString(pDict, "menuItemClicked");

        if (pFunc && PyCallable_Check(pFunc))
        {
            pArgs = PyTuple_New(3);

            pValue = PyInt_FromLong((int)this);
            PyTuple_SetItem(pArgs, 0, pValue);

            int menu = 0;
            for(int i = 0; i < menuList->count(); i++)
            {
                KPopupMenu* tmp;
                if(i==0)
                {
                    tmp = (KPopupMenu*) menuList->first();
                }
                else
                {
                    tmp = (KPopupMenu*) menuList->next();
                }
                if(tmp != 0)
                {
                    if(tmp->isItemVisible(id))
                    {
                        menu = (int) tmp;
                        break;
                    }
                }
            }

            pValue = PyInt_FromLong(menu);
            PyTuple_SetItem(pArgs, 1, pValue);

            pValue = PyInt_FromLong(id);
            PyTuple_SetItem(pArgs, 2, pValue);


            pValue = PyObject_CallObject(pFunc, pArgs);
            if (pValue != NULL)
            {
                Py_DECREF(pValue);
            }
            else
            {
                qWarning("Call to menuItemClicked failed");
		PyErr_Print();
            }
            Py_DECREF(pArgs);
        }
        else
        {
	  //can't find callback
        }

        // swap my thread state out of the interpreter
        PyThreadState_Swap(NULL);
        // clear out any cruft from thread state object
        PyThreadState_Clear(myThreadState);
        // delete my thread state object
        PyThreadState_Delete(myThreadState);
        // release the lock
        PyEval_ReleaseLock();
    }
}

void karamba::activeTaskChanged(Task* t)
{
    //Everything below is to call the python callback function

    if (pythonThemeExtensionLoaded)
    {

        // get the global lock
        PyEval_AcquireLock();
        // get a reference to the PyInterpreterState

        //Python thread references
        extern PyThreadState* mainThreadState;

        PyInterpreterState* mainInterpreterState = mainThreadState -> interp;
        // create a thread state object for this thread
        PyThreadState* myThreadState = PyThreadState_New(mainInterpreterState);
        PyThreadState_Swap(myThreadState);

        pFunc = PyDict_GetItemString(pDict, "activeTaskChanged");

        if (pFunc && PyCallable_Check(pFunc))
        {
            pArgs = PyTuple_New(2);

            pValue = PyInt_FromLong((int) this);
            PyTuple_SetItem(pArgs, 0, pValue);

            pValue = PyInt_FromLong((int) t);
            PyTuple_SetItem(pArgs, 1, pValue);

            pValue = PyObject_CallObject(pFunc, pArgs);
            if (pValue != NULL)
            {
                Py_DECREF(pValue);
            }
            else
            {
                qWarning("Call to activeTaskChanged failed");
		PyErr_Print();
            }
            Py_DECREF(pArgs);
        }
        else
        {
	  //callback not found
        }

        // swap my thread state out of the interpreter
        PyThreadState_Swap(NULL);
        // clear out any cruft from thread state object
        PyThreadState_Clear(myThreadState);
        // delete my thread state object
        PyThreadState_Delete(myThreadState);
        // release the lock
        PyEval_ReleaseLock();
    }
}

void karamba::taskAdded(Task* t)
{
    //Everything below is to call the python callback function

    if (pythonThemeExtensionLoaded)
    {

        // get the global lock
        PyEval_AcquireLock();
        // get a reference to the PyInterpreterState

        //Python thread references
        extern PyThreadState* mainThreadState;

        PyInterpreterState* mainInterpreterState = mainThreadState -> interp;
        // create a thread state object for this thread
        PyThreadState* myThreadState = PyThreadState_New(mainInterpreterState);
        PyThreadState_Swap(myThreadState);

        pFunc = PyDict_GetItemString(pDict, "taskAdded");

        if (pFunc && PyCallable_Check(pFunc))
        {
            pArgs = PyTuple_New(2);

            pValue = PyInt_FromLong((int) this);
            PyTuple_SetItem(pArgs, 0, pValue);

            pValue = PyInt_FromLong((int) t);
            PyTuple_SetItem(pArgs, 1, pValue);

            pValue = PyObject_CallObject(pFunc, pArgs);
            if (pValue != NULL)
            {
                Py_DECREF(pValue);
            }
            else
            {
                qWarning("Call to taskAdded failed");
		PyErr_Print();
            }
            Py_DECREF(pArgs);
        }
        else
        {
	  //callback not found
        }

        // swap my thread state out of the interpreter
        PyThreadState_Swap(NULL);
        // clear out any cruft from thread state object
        PyThreadState_Clear(myThreadState);
        // delete my thread state object
        PyThreadState_Delete(myThreadState);
        // release the lock
        PyEval_ReleaseLock();
    }
}

void karamba::taskRemoved(Task* t)
{
    //Everything below is to call the python callback function

    if (pythonThemeExtensionLoaded)
    {

        // get the global lock
        PyEval_AcquireLock();
        // get a reference to the PyInterpreterState

        //Python thread references
        extern PyThreadState* mainThreadState;

        PyInterpreterState* mainInterpreterState = mainThreadState -> interp;
        // create a thread state object for this thread
        PyThreadState* myThreadState = PyThreadState_New(mainInterpreterState);
        PyThreadState_Swap(myThreadState);

        pFunc = PyDict_GetItemString(pDict, "taskRemoved");

        if (pFunc && PyCallable_Check(pFunc))
        {
            pArgs = PyTuple_New(2);

            pValue = PyInt_FromLong((int) this);
            PyTuple_SetItem(pArgs, 0, pValue);

            pValue = PyInt_FromLong((int) t);
            PyTuple_SetItem(pArgs, 1, pValue);

            pValue = PyObject_CallObject(pFunc, pArgs);
            if (pValue != NULL)
            {
                Py_DECREF(pValue);
            }
            else
            {
                qWarning("Call to taskRemoved failed");
		PyErr_Print();
            }
            Py_DECREF(pArgs);
        }
        else
        {
	  //callback not found
        }

        // swap my thread state out of the interpreter
        PyThreadState_Swap(NULL);
        // clear out any cruft from thread state object
        PyThreadState_Clear(myThreadState);
        // delete my thread state object
        PyThreadState_Delete(myThreadState);
        // release the lock
        PyEval_ReleaseLock();
    }
}

void karamba::startupAdded(Startup* t)
{
    //Everything below is to call the python callback function

    if (pythonThemeExtensionLoaded)
    {

        // get the global lock
        PyEval_AcquireLock();
        // get a reference to the PyInterpreterState

        //Python thread references
        extern PyThreadState* mainThreadState;

        PyInterpreterState* mainInterpreterState = mainThreadState -> interp;
        // create a thread state object for this thread
        PyThreadState* myThreadState = PyThreadState_New(mainInterpreterState);
        PyThreadState_Swap(myThreadState);

        pFunc = PyDict_GetItemString(pDict, "startupAdded");

        if (pFunc && PyCallable_Check(pFunc))
        {
            pArgs = PyTuple_New(2);

            pValue = PyInt_FromLong((int) this);
            PyTuple_SetItem(pArgs, 0, pValue);

            pValue = PyInt_FromLong((int) t);
            PyTuple_SetItem(pArgs, 1, pValue);

            pValue = PyObject_CallObject(pFunc, pArgs);
            if (pValue != NULL)
            {
                Py_DECREF(pValue);
            }
            else
            {
                qWarning("Call to startupAdded failed");
		PyErr_Print();
            }
            Py_DECREF(pArgs);
        }
        else
        {
	  //callback not found
        }

        // swap my thread state out of the interpreter
        PyThreadState_Swap(NULL);
        // clear out any cruft from thread state object
        PyThreadState_Clear(myThreadState);
        // delete my thread state object
        PyThreadState_Delete(myThreadState);
        // release the lock
        PyEval_ReleaseLock();
    }
}

void karamba::startupRemoved(Startup* t)
{
    //Everything below is to call the python callback function

    if (pythonThemeExtensionLoaded)
    {

        // get the global lock
        PyEval_AcquireLock();
        // get a reference to the PyInterpreterState

        //Python thread references
        extern PyThreadState* mainThreadState;

        PyInterpreterState* mainInterpreterState = mainThreadState -> interp;
        // create a thread state object for this thread
        PyThreadState* myThreadState = PyThreadState_New(mainInterpreterState);
        PyThreadState_Swap(myThreadState);

        pFunc = PyDict_GetItemString(pDict, "startupRemoved");

        if (pFunc && PyCallable_Check(pFunc))
        {
            pArgs = PyTuple_New(2);

            pValue = PyInt_FromLong((int) this);
            PyTuple_SetItem(pArgs, 0, pValue);

            pValue = PyInt_FromLong((int) t);
            PyTuple_SetItem(pArgs, 1, pValue);

            pValue = PyObject_CallObject(pFunc, pArgs);
            if (pValue != NULL)
            {
                Py_DECREF(pValue);
            }
            else
            {
                qWarning("Call to startupRemoved failed");
		PyErr_Print();
            }
            Py_DECREF(pArgs);
        }
        else
        {
	  //cannot find callback
        }

        // swap my thread state out of the interpreter
        PyThreadState_Swap(NULL);
        // clear out any cruft from thread state object
        PyThreadState_Clear(myThreadState);
        // delete my thread state object
        PyThreadState_Delete(myThreadState);
        // release the lock
        PyEval_ReleaseLock();
    }
}

void  karamba::processExited (KProcess *proc) {
  //Nothing needed here at the moment.
  //Could add a "commandFinished" callback here if it becomes needed at some point.
}

void  karamba::receivedStdout (KProcess *proc, char *buffer, int buflen) {

  //Everything below is to call the python callback function

  if (pythonThemeExtensionLoaded) {

    // get the global lock
    PyEval_AcquireLock();
    // get a reference to the PyInterpreterState

    //Python thread references
    extern PyThreadState * mainThreadState;

    PyInterpreterState * mainInterpreterState = mainThreadState->interp;
    // create a thread state object for this thread
    PyThreadState * myThreadState = PyThreadState_New(mainInterpreterState);
    PyThreadState_Swap(myThreadState);

    pFunc = PyDict_GetItemString(pDict, "commandOutput");

    if (pFunc && PyCallable_Check(pFunc)) {
      pArgs = PyTuple_New(3);

      pValue = PyInt_FromLong((int)this);
      PyTuple_SetItem(pArgs, 0, pValue);

      pValue = PyInt_FromLong((int)proc->pid());
      PyTuple_SetItem(pArgs, 1, pValue);

      pValue = PyString_FromString(buffer);
      PyTuple_SetItem(pArgs, 2, pValue);

      pValue = PyObject_CallObject(pFunc, pArgs);
      if (pValue != NULL) {
	Py_DECREF(pValue);
      }
      else {
	qWarning("Call to commandOutput failed");
	PyErr_Print();
      }
      Py_DECREF(pArgs);
      }
    else {
      //callback not found
    }


    // swap my thread state out of the interpreter
    PyThreadState_Swap(NULL);
    // clear out any cruft from thread state object
    PyThreadState_Clear(myThreadState);
    // delete my thread state object
    PyThreadState_Delete(myThreadState);
    // release the lock
    PyEval_ReleaseLock();
  }

}


//For KDE session management
void karamba::saveProperties(KConfig* config) {
  config->setGroup("session");
  config->writeEntry("theme", themeFile);
  writeConfigData();
}

//For KDE session management
void karamba::readProperties(KConfig* config) {
  config->setGroup("session");
  QString atheme = config->readEntry("theme","");
}

//Register types of events that can be dragged on our widget
void karamba::dragEnterEvent(QDragEnterEvent* event)
{
    event->accept(
        QTextDrag::canDecode(event)
    );
}

//Handle the drop part of a drag and drop event.
void karamba::dropEvent(QDropEvent* event)
{
    QString text;

    if ( QTextDrag::decode(event, text) ) {

      //Everything below is to call the python callback function

      if (pythonThemeExtensionLoaded) {

	// get the global lock
	PyEval_AcquireLock();
	// get a reference to the PyInterpreterState

	//Python thread references
	extern PyThreadState * mainThreadState;

	PyInterpreterState * mainInterpreterState = mainThreadState->interp;
	// create a thread state object for this thread
	PyThreadState * myThreadState = PyThreadState_New(mainInterpreterState);
	PyThreadState_Swap(myThreadState);

	pFunc = PyDict_GetItemString(pDict, "itemDropped");

	if (pFunc && PyCallable_Check(pFunc)) {
	  pArgs = PyTuple_New(2);

	  pValue = PyInt_FromLong((int)this);
	  PyTuple_SetItem(pArgs, 0, pValue);

	  pValue = PyString_FromString(text.ascii());
	  PyTuple_SetItem(pArgs, 1, pValue);

	  pValue = PyObject_CallObject(pFunc, pArgs);
	  if (pValue != NULL) {
	    Py_DECREF(pValue);
	  }
	  else {
	    qWarning("Call to itemDropped failed");
	    PyErr_Print();
	  }
	  Py_DECREF(pArgs);
	}
	else {
	  //callback failed
	}


	// swap my thread state out of the interpreter
	PyThreadState_Swap(NULL);
	// clear out any cruft from thread state object
	PyThreadState_Clear(myThreadState);
	// delete my thread state object
	PyThreadState_Delete(myThreadState);
	// release the lock
	PyEval_ReleaseLock();
      }

    }
}

void karamba::toDesktop(int id, int menuid)
{
    int i;

    desktop = id;
    for (i=0; ; i++)
    {
        int mid = toDesktopMenu->idAt(i);
        if (mid == -1)
            break;

        toDesktopMenu->setItemChecked(mid, false);
    }
    toDesktopMenu->setItemChecked(menuid, true);

    if (desktop)
        info->setDesktop( desktop);
    else
        info->setDesktop( NETWinInfo::OnAllDesktops );
}

//Eventually maybe add a theme browser here
void karamba::downloadThemes() {
  //Go to SK Themes webpage for now
  KRun::runURL( KURL( "http://www.superkaramba.com/" ), "text/html" );
}


SignalBridge::SignalBridge(QObject* parent, QString name, KActionCollection* ac) : QObject(parent, name),
   collection(ac)
{
  setName(name);
}

void SignalBridge::receive()
{
  emit enabled(name(), ((KToggleAction*)collection -> action(name())) -> isChecked());
}

DesktopChangeSlot::DesktopChangeSlot(QObject *parent, int id)
    : QObject(parent, "")
{
    desktopid = id;
}

void DesktopChangeSlot::receive()
{
    karamba *k = (karamba *)parent();

    /* XXX - check type cast */

    k->toDesktop(desktopid, menuid);
}

void DesktopChangeSlot::setMenuId(int id)
{
    menuid = id;
}

int DesktopChangeSlot::menuId()
{
    return menuid;
}

#include "karamba.moc"
